':' ''' '
autocomplete() {
    if [ $# != 2 -o ! -f "$2" ]
    then
        echo "USAGE: autocomplete <program-name> <commands-path>"
        return 1
    fi
    source /dev/stdin <<EOF
    _$1() {
        COMP_WORDBREAKS=" \"'@><=;|&(:,"
        IFS=$'\\n' read -rd '' -a RESULT <<< "\$(python -m autocompleter "$(readlink -e $2)" "\$COMP_LINE" "\$COMP_POINT")"
        COMPREPLY=("\${RESULT[@]:1}")
        eval "\${RESULT[0]}"
    }
    complete -F _$1 $1
EOF
}
return
'''

__version__ = 0, 1, 0

import imp, os, re, shlex, sys, traceback

completers   = {}
post_command = 'compopt -o nospace'

def autocomplete(commands, line, cursor):
    commands                       = parse_commands(commands)
    words, cword, word, key, value = parse_line(line, cursor)
    if cword == 1:
        return complete_command(commands, word)
    command = commands.get(words[1])
    if not command:
        return []
    if value is None:
        args = parse_items(words[2:], ignore=[word])
        return complete_key(command, args, key)
    else:
        if key not in command:
            return []
        return complete_value(command[key], value)

def parse_commands(commands):
    result = {}
    for command in commands.strip().split(os.linesep):
        if ' ' in command:
            command, args   = command.split(' ', 1)
            result[command] = parse_items(shlex.split(args))
        else:
            result[command] = {}
    return result

def parse_items(items, ignore=[]):
    result = {}
    for item in items:
        if item in ignore:
            continue
        if '=' in item:
            key, value = item.split('=', 1)
        else:
            key, value = item, None
        result[key] = value
    return result

def parse_line(line, cursor):
    words = []
    cword = 0
    lexer = shlex.shlex(line)
    lexer.whitespace_split = True
    for n, word in enumerate(lexer):
        words.append(word)
        if not cword and lexer.instream.tell() >= cursor:
            cword = n
    if lexer.instream.tell() == cursor and line[cursor-1] == ' ':
        cword += 1
    if len(words) > cword:
        word = words[cword]
        if '=' in word:
            key, value = word.split('=', 1)
        else:
            key, value = word, None
    else:
        word, key, value = '', '', None
    return words, cword, word, key, value

def complete_command(commands, word):
    return [command+' ' for command in commands if command.startswith(word)]

def complete_key(command, args, key):
    completions = []
    for arg in command:
        if arg.startswith(key) and arg not in args:
            if command[arg]:
                completions.append(arg+'=')
            else:
                completions.append(arg+' ')
    return completions

def complete_value(pattern, value):
    for name, completer in completers.items():
        match = re.match(name, pattern)
        if match:
            return completer(value, *match.groups())
    return []

def complete(pattern):
    def decorator(function):
        completers[pattern] = function
        return function
    return decorator

@complete('<path>')
def path_completer(value):
    global post_command
    post_command = 'compopt -o filenames'
    directory, filename = os.path.split(value)
    if not directory:
        directory = '.'
    if not os.path.exists(directory):
        return []
    completions = []
    for entry in os.listdir(directory):
        if entry.startswith(filename):
            if directory != '.':
                entry = os.path.join(directory, entry)
            completions.append(entry)
    return completions

@complete('<option:(.*)>')
def option_completer(value, options):
    return [option+' ' for option in options.split('|') if option.startswith(value)]

def main(commands_path, line, cursor):
    try:
        sys.modules[os.path.splitext(os.path.basename(__file__))[0]] = sys.modules[__name__]
        commands = imp.load_source('commands', commands_path).__doc__
    except:
        commands = open(commands_path).read()
    completions = autocomplete(commands, line, cursor)
    print(post_command + '\n' + '\n'.join(completions))

if __name__ == '__main__':
    if len(sys.argv) != 4:
        print('USAGE: %s <configuration-path> <line> <cursor>' % sys.argv[0])
    try:
        main(commands_path=sys.argv[1], line=sys.argv[2], cursor=int(sys.argv[3]))
    except:
        traceback.print_exc()
